#include "Compiler.h"
#include "codeGen/Quadruple.h"
#include "codeGen/RegisterAllocator.h"
#include <filesystem>
#include <iostream>
#include <cassert>

extern std::filesystem::path binaryPath;
const std::string Compiler::extension = "s";
void Compiler::externalCommand(std::filesystem::path lat) {
	auto obj = lat;
	obj.replace_extension("o");
	auto cmd = std::string("as --32 '") + lat.string() + "' -o '" + obj.string() + "'";
	cerr << cmd << endl;
	system(cmd.data());
	auto lib = binaryPath.parent_path() / "lib";
	auto runtime = lib / "runtime.o";
	auto result = obj;
	result.replace_extension("");
	cmd = std::string("ld -m elf_i386 -L'") + lib.string() + "' '" + runtime.string() + "' '" + obj.string() + "' -o '" + result.string() + "'";
	cerr << cmd << endl;
	system(cmd.data());
}

string Compiler::compile(Visitable *v) {
	buf <<
		".globl _start\n"
	    ".globl exit\n"
		"_start:\n"
		"	pushl %ebp\n"
		"	movl %esp, %ebp\n"
		"	call main\n"
		"	pushl %eax\n"
		"	call exit\n";

	for (const auto& c : scope->classes) {
		c->calculateSize();
	}

	for (const auto& f : scope->functions) {
		compileFunction(f);
	}

	return buf.str();
}

void Compiler::compileFunction(const FunctionInfoPtr& f) {
	string name = mangleFunctionName(f);
	buf << ".globl \"" << name << "\"\n";

	if (!f->block) return;

	// TODO: jeśli funkcja jest klasowa, ustawić self na ... wartosc z ecx
	// c++ thiscall ecx, stack right-to-left (push c, push b, push a)

	buf << "\"" << name << "\":\n";

	auto quadEnv = quadGen.compileFunction(f);

	RegisterAllocator regGen = RegisterAllocator(quadEnv);

	regGen.allocate();

	int blkNo = 0;
	for (const auto& b : quadEnv.blocks) {
		buf << "------------------------------\n";
		buf << "blok " << blkNo << " " << b->getName() << endl;
		/*buf << "in: ";
		for (auto in : b->in) {
			buf << in->getName() << " ";
		}
		buf << "\nouts: ";
		for (auto out : b->out) {
			buf << out->getName() << " ";
		}
		buf << "\n";*/
		if (!b->flow.in.empty()) {
			buf << "in: ";
			for (const auto& v : b->flow.in) {
				buf << v->name << ", ";
			}
			buf << "\n----------\n";
		}
		if (!b->flow.def.empty()) {
			buf << "def: ";
			for (const auto& v : b->flow.def) {
				buf << v->name << ", ";
			}
			buf << "\n----------\n";
		}
		for (const auto& q : b->quads) {
			auto ret = q->toString();
			if (!ret.empty()) {
				buf << ret << " alive: ";
				for (auto var : q->aliveAfter)
					buf << var->name << " ";
				buf << endl;
			}
		}
		if (!b->flow.use.empty()) {
			buf << "----------\nuse: ";
			for (const auto &v : b->flow.use) {
				buf << v->name << ", ";
			}
			buf << "\n";
		}
		if (!b->flow.out.empty()) {
			buf << "----------\nout: ";
			for (const auto &v : b->flow.out) {
				buf << v->name << ", ";
			}
			buf << "\n";
		}
		blkNo++;
	}
}
