#include "RegisterAllocator.h"
#include "BasicBlock.h"
#include "setOverloads.h"

void RegisterAllocator::analyseLive() {
	for (const auto& v : vars) {
		if (!v->info) continue;
		for (const auto& q : v->writes) {
			q->block->flow.addDef(v);
		}
		for (const auto& q : v->uses) {
			q->block->flow.addUse(v);
		}
	}

	for (const auto& b : blocks) {
		for (const auto& phiMap : b->phi->phi) {
			for (const auto& phiRow : phiMap) {
				b->flow.def.emplace(phiRow.first);
			}
		}
	}

	while (true) {
		bool changed = false;
		for (const auto& b : blocks) {
			set<VariablePtr> pIn, pOut, diff;
			auto &f = b->flow;
			pIn.swap(f.in);
			pOut.swap(f.out);
			f.in = f.use + (pOut - f.def);
			for (const auto& succ : b->out) {
				f.out += succ->flow.inForBlock(b);
			}
			if (f.in != pIn || f.out != pOut) changed = true;
		}
		if (!changed) break;
	}

	assert(blocks.empty() || blocks[0]->flow.in.empty());
}

void RegisterAllocator::allocate() {
	analyseLive();
	buildGraph();
}

void RegisterAllocator::buildGraph() {
	for (const auto& b : blocks) {
		set<VariablePtr> alive = b->flow.in;
		for (const auto& q : b->quads) {
			alive += to_set(q->definitions());
			vector<VariablePtr> dead;
			for (const auto& var : alive) {
				// this quad is the last quad that uses var
				if ((var->uses.empty() || var->uses.back() == q) && !b->flow.out.contains(var)) {
					dead.emplace_back(var);
				}
			}
			alive -= to_set(dead);
			q->aliveAfter = alive;
			for (auto i : alive) {
				for (auto j : alive) {
					if (i != j) {
						if (i.get() > j.get()) {
							swap(i, j);
						}
						graphEdges.insert(make_pair(i, j));
					}
				}
			}
		}
	}
}


